"""
Tests for `typed_settings.attrs.hooks`.
"""
import functools
import json
from datetime import datetime
from enum import Enum
from typing import Any, Dict, List, Optional, Set, Tuple, Type, Union

import attr
import pytest

from typed_settings.attrs.hooks import auto_convert, auto_serialize
from typed_settings.attrs.validators import le


auto_converter = functools.partial(attr.frozen, field_transformer=auto_convert)


class TestTransformHook:
    """
    Tests for `attrs(tranform_value_serializer=func)`
    """

    def test_hook_applied(self):
        """
        The transform hook is applied to all attributes.  Types can be missing,
        explicitly set, or annotated.
        """
        results = []

        def hook(cls, attribs):
            results[:] = [(a.name, a.type) for a in attribs]
            return attribs

        @attr.s(field_transformer=hook)
        class C:
            x = attr.ib()
            y = attr.ib(type=int)
            z: float = attr.ib()

        assert results == [("x", None), ("y", int), ("z", float)]

    def test_hook_applied_auto_attrib(self):
        """
        The transform hook is applied to all attributes and type annotations
        are detected.
        """
        results = []

        def hook(cls, attribs):
            results[:] = [(a.name, a.type) for a in attribs]
            return attribs

        @attr.s(auto_attribs=True, field_transformer=hook)
        class C:
            x: int
            y: str = attr.ib()

        assert results == [("x", int), ("y", str)]

    def test_hook_applied_modify_attrib(self):
        """
        The transform hook can modify attributes.
        """

        def hook(cls, attribs):
            return [a.evolve(converter=a.type) for a in attribs]

        @attr.s(auto_attribs=True, field_transformer=hook)
        class C:
            x: int = attr.ib(converter=int)
            y: float

        c = C(x="3", y="3.14")
        assert c == C(x=3, y=3.14)

    def test_hook_remove_field(self):
        """
        It is possible to remove fields via the hook.
        """

        def hook(cls, attribs):
            return [a for a in attribs if a.type is not int]

        @attr.s(auto_attribs=True, field_transformer=hook)
        class C:
            x: int
            y: float

        assert attr.asdict(C(2.7)) == {"y": 2.7}

    def test_hook_add_field(self):
        """
        It is possible to add fields via the hook.
        """

        def hook(cls, attribs):
            a1 = attribs[0]
            a2 = a1.evolve(name="new")
            return [a1, a2]

        @attr.s(auto_attribs=True, field_transformer=hook)
        class C:
            x: int

        assert attr.asdict(C(1, 2)) == {"x": 1, "new": 2}

    def test_hook_with_inheritance(self):
        """
        The hook receives all fields from base classes.
        """

        def hook(cls, attribs):
            assert [a.name for a in attribs] == ["x", "y"]
            # Remove Base' "x"
            return attribs[1:]

        @attr.s(auto_attribs=True)
        class Base:
            x: int

        @attr.s(auto_attribs=True, field_transformer=hook)
        class Sub(Base):
            y: int

        assert attr.asdict(Sub(2)) == {"y": 2}


class TestAsDictHook:
    def test_asdict(self):
        """
        asdict() calls the hooks in attrs classes and in other datastructures
        like lists or dicts.
        """

        def hook(inst, a, v):
            if isinstance(v, datetime):
                return v.isoformat()
            return v

        @attr.dataclass
        class Child:
            x: datetime
            y: List[datetime]

        @attr.dataclass
        class Parent:
            a: Child
            b: List[Child]
            c: Dict[str, Child]
            d: Dict[str, datetime]

        inst = Parent(
            a=Child(1, [datetime(2020, 7, 1)]),
            b=[Child(2, [datetime(2020, 7, 2)])],
            c={"spam": Child(3, [datetime(2020, 7, 3)])},
            d={"eggs": datetime(2020, 7, 4)},
        )

        result = attr.asdict(inst, value_serializer=hook)
        assert result == {
            "a": {"x": 1, "y": ["2020-07-01T00:00:00"]},
            "b": [{"x": 2, "y": ["2020-07-02T00:00:00"]}],
            "c": {"spam": {"x": 3, "y": ["2020-07-03T00:00:00"]}},
            "d": {"eggs": "2020-07-04T00:00:00"},
        }

    def test_asdict_calls(self):
        """
        The correct instances and attribute names are passed to the hook.
        """
        calls = []

        def hook(inst, a, v):
            calls.append((inst, a.name if a else a, v))
            return v

        @attr.dataclass
        class Child:
            x: int

        @attr.dataclass
        class Parent:
            a: Child
            b: List[Child]
            c: Dict[str, Child]

        inst = Parent(a=Child(1), b=[Child(2)], c={"spam": Child(3)})

        attr.asdict(inst, value_serializer=hook)
        assert calls == [
            (inst, "a", inst.a),
            (inst.a, "x", inst.a.x),
            (inst, "b", inst.b),
            (inst.b[0], "x", inst.b[0].x),
            (inst, "c", inst.c),
            (None, None, "spam"),
            (inst.c["spam"], "x", inst.c["spam"].x),
        ]


class LeEnum(Enum):
    spam = "Le spam"
    eggs = "Le eggs"


@auto_converter
class Child:
    x: int = attr.ib()
    y: int = attr.ib(converter=float)


@auto_converter(kw_only=True)
class Parent:
    child: Child
    a: float
    b: float = attr.field(default=3.14, validator=le(2))
    c: LeEnum
    d: datetime
    e: "List[Child]"
    f: Set[datetime]


class TestAutoConvertHook:
    """Tests for the bundled auto-convert hook."""

    DATA = {
        "a": "1",
        "b": "2",
        "c": "Le spam",
        "d": "2020-05-04T13:37:00",
        "e": [{"x": "23", "y": "42"}],
        "f": ["2020-05-04T13:37:00", "2020-05-04T13:37:00"],
        "child": {"x": 23, "y": "42"},  # Also tests int->str conversion
    }

    @pytest.fixture(scope="class")
    def parent(self):
        return Parent(**self.DATA)

    def test_convert_to_parent(self, parent):
        """
        The auto_convert hook must convert attrs classes, datetimes, enums and
        basic type as well as basic containers.
        """
        assert parent == Parent(
            a=1.0,
            b=2.0,
            c=LeEnum.spam,
            d=datetime(2020, 5, 4, 13, 37),
            e=[Child(23, 42)],
            f={datetime(2020, 5, 4, 13, 37)},
            child=Child(23, 42),
        )

    def test_serialize_from_parent(self, parent):
        """
        The serialize_hook must be able to serialize the same types as the
        auto_convert hook.  The set with duplicate entries of attrib "f" is
        not the same as in the original dict!
        """
        d = attr.asdict(parent, value_serializer=auto_serialize)
        assert d == {
            "a": 1.0,
            "b": 2.0,
            "c": "Le spam",
            "d": "2020-05-04T13:37:00",
            "e": [{"x": 23, "y": 42}],
            "f": ["2020-05-04T13:37:00"],
            "child": {"x": 23, "y": 42},
        }

    def test_json_roundtrip(self, parent):
        """
        The roundtrip "inst -> JSON -> inst" must result in the same inst.
        """
        d = attr.asdict(parent, value_serializer=auto_serialize)
        assert Parent(**json.loads(json.dumps(d))) == parent

    def test_convert_to_struct_tuple(self):
        """
        Tuples can be defined like structs.
        """

        @auto_converter
        class C:
            x: Tuple[int, float, str]

        c = C(["3", "3.2", "3.2"])
        assert c.x == (3, 3.2, "3.2")

    def test_convert_to_iterable_tuple(self):
        """
        Tuples can be defined like (immutable) lists.
        """

        @auto_converter
        class C:
            x: Tuple[int, ...]

        c = C(["3", "2", "1"])
        assert c.x == (3, 2, 1)

    def test_to_mapping(self):
        """
        Converters are applied to dict keys and values.
        """

        @auto_converter
        class C:
            x: Dict[int, float]

        c = C({"1": "2"})
        assert c.x == {1: 2.0}

    @pytest.mark.parametrize("val, expected", [({}, None), ({"x": "2"}, 2)])
    def test_convert_to_optional(self, val, expected):
        """
        Conversion to Optional works with and without a value.
        """

        @auto_converter
        class C:
            x: Optional[int] = None

        c = C(**val)
        assert c.x == expected
        assert type(c.x) is type(expected)

    def test_convert_to_union(self):
        """
        Union values resolve to the first matching type
        """

        @auto_converter
        class C:
            x: Union[int, float]

        c = C(**{"x": "3.2"})
        assert c.x == 3.2
        assert type(c.x) is float

    def test_invalid_generic_type(self):
        """
        Annotating a generic type that the converter doesn't know leads to
        a TypeError.
        """
        with pytest.raises(
            TypeError, match="Cannot create converter for generic type:"
        ):

            @auto_converter
            class C:
                x: Type[int]

    def test_nested_conversion(self):
        """
        Generics (like lists and dicts) can be nested and will still be
        properly converted.
        """

        @auto_converter
        class A:
            x: int
            y: int

        @auto_converter
        class C:
            x: Dict[Tuple[int, int], List[Dict[int, A]]]

        c = C(
            {
                ("1", "2"): [
                    {"3": {"x": "4", "y": "5"}},
                    {"6": {"x": "7", "y": "8"}},
                ],
            }
        )
        assert c.x == {(1, 2): [{3: A(4, 5)}, {6: A(7, 8)}]}

    def test_any(self):
        """
        "Any" don't convert anything but also do not cause any problems.
        """

        @auto_converter
        class C:
            x: Any
            y: List[Any]

        o = object()
        c = C(23, [True, None, 23, 3.14, "spam", o])
        assert c.x == 23
        assert c.y == [True, None, 23, 3.14, "spam", o]



"""
Additional attrs converters
"""
from datetime import datetime


__all__ = [
    "to_attrs",
    "to_dt",
    "to_iterable",
    "to_mapping",
    "to_tuple",
    "to_union",
]


def to_attrs(cls):
    """
    A converter that creates an instance of *cls* from a dict but leaves
    instances of that class as they are.

    Classes can define a ``from_dict()`` classmethod which will be called
    instead of the their `__init__()`.  This can be useful if you want to
    create different sub classes of *cls* depending on the data (e.g.,
    a ``Cat`` or a ``Dog`` inheriting ``Animal``).

    :param type cls: The class to convert data to.
    :returns: The converter function for *cls*.
    :rtype: callable

    .. versionadded:: 20.3.0

    """
    type_ = cls.from_dict if hasattr(cls, "from_dict") else cls

    def convert(val):
        if not isinstance(val, (cls, dict)):
            raise TypeError(
                f'Invalid type "{type(val).__name__}"; expected '
                f'"{cls.__name__}" or "dict".'
            )
        return type_(**val) if isinstance(val, dict) else val

    n = cls.__name__
    convert.__doc__ = f"""
        Convert *data* to an intance of {n} if it is not already an instance
        of it.

        :param Union[dict, {n}] data: The input data
        :returns: The converted data
        :rtype: {n}
        :raises TypeError: if *data* is neither a dict nor an instance of {n}.
        """

    return convert


def to_dt(val):
    """
    Convert an ISO formatted string to :class:`datetime.datetime`.  Leave the
    input untouched if it is already a datetime.

    See: :func:`datetime.datetime.fromisoformat`

    :param Union[str, datetime.datetime] data: The input data
    :returns: A parsed datetime object
    :rtype: datetime.datetime
    :raises TypeError: If *val* is neither a str nor a datetime.

    .. versionadded:: 20.3.0
    """
    if not isinstance(val, (datetime, str)):
        raise TypeError(
            f'Invalid type "{type(val).__name__}"; expected "datetime" or '
            f'"str".'
        )
    if isinstance(val, str):
        if val[-1] == "Z":
            val = val.replace("Z", "+00:00")
        return datetime.fromisoformat(val)
    return val


def to_bool(val):
    """
    Convert "boolean" strings (e.g., from env. vars.) to real booleans.

    Values mapping to :code:`True`:

    - :code:`True`
    - :code:`"True"`
    - :code:`"true"`
    - :code:`"yes"`
    - :code:`"1"`
    - :code:`1`

    Values mapping to :code:`False`:

    - :code:`False`
    - :code:`"False"`
    - :code:`"false"`
    - :code:`"no"`
    - :code:`"0"`
    - :code:`0`

    Raise :exc:`ValueError` for any other value.
    """
    truthy = {True, "True", "true", "yes", "1", 1}
    falsy = {False, "False", "false", "no", "0", 0}
    try:
        if val in truthy:
            return True
        if val in falsy:
            return False
    except TypeError:
        # Raised when "val" is not hashable (e.g., lists)
        pass
    raise ValueError(f"Cannot convert value to bool: {val}")


def to_iterable(cls, converter):
    """
    A converter that creates a *cls* iterable (e.g., ``list``) and calls
    *converter* for each element.

    :param Type[Iterable] cls: The type of the iterable to create
    :param callable converter: The converter to apply to all items of the
        input data.
    :returns: The converter function
    :rtype: callable
    """

    def convert(val):
        return cls(converter(d) for d in val)

    return convert


def to_tuple(cls, converters):
    """
    A converter that creates a struct-like tuple (or namedtuple or similar)
    and converts each item via the corresponding converter from *converters*

    The input value must have exactly as many elements as there are converters.

    :param Type[Tuple] cls: The type of the tuple to create
    :param List[callable] converters: The respective converters for each tuple
        item.
    :returns: The converter function
    :rtype: callable
    """

    def convert(val):
        if len(val) != len(converters):
            raise TypeError(
                "Value must have {} items but has: {}".format(
                    len(converters), len(val)
                )
            )
        return cls(c(v) for c, v in zip(converters, val))

    return convert


def to_mapping(cls, key_converter, val_converter):
    """
    A converter that creates a mapping and converts all keys and values using
    the respective converters.

    :param Type[Mapping] cls: The mapping type to create (e.g., ``dict``).
    :param callable key_converter: The converter function to apply to all keys.
    :param callable val_converter: The converter function to apply to all
        values.
    :returns: The converter function
    :rtype: callable
    """

    def convert(val):
        return cls(
            (key_converter(k), val_converter(v)) for k, v in val.items()
        )

    return convert


def to_union(converters):
    """
    A converter that applies a number of converters to the input value and
    returns the result of the first converter that does not raise a
    :exc:`TypeError` or :exc:`ValueError`.

    If the input value already has one of the required types, it will be
    returned unchanged.

    :param List[callable] converters: A list of converters to try on the input.
    :returns: The converter function
    :rtype: callable

    """

    def convert(val):
        if type(val) in converters:
            # Preserve val as-is if it already has a matching type.
            # Otherwise float(3.2) would be converted to int
            # if the converters are [int, float].
            return val
        for converter in converters:
            try:
                return converter(val)
            except (TypeError, ValueError):
                pass
        raise ValueError(
            "Failed to convert value to any Union type: {}".format(val)
        )

    return convert



# -- Path setup --------------------------------------------------------------

import re
from pathlib import Path


def find_version():
    path = Path(__file__).parent.parent.joinpath("setup.py")
    version_file = path.read_text()
    version_match = re.search(r'^ +version="([^"]+)",$', version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string")


# -- Project information -----------------------------------------------------


project = "Typed Settings"
author = "Stefan Scherfke"
copyright = "2020, Stefan Scherfke"
release = find_version()
version = ".".join(release.split(".")[0:2])


# -- General configuration ---------------------------------------------------

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx_autodoc_typehints",
]
templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------

html_theme = "furo"
# html_theme_options = {
#     "logo_only": True,
#     "sidebar_hide_name": True,
# }

html_static_path = ["_static"]
html_logo = "_static/typed-settings-spacing.svg"
html_title = "Typed Settings"


# -- Extension configuration -------------------------------------------------

# Autodoc
autodoc_member_order = "bysource"

# Intersphinx
intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "attrs": ("https://www.attrs.org/en/stable/", None),
    "click": ("https://click.palletsprojects.com/en/master/", None),
}



from datetime import datetime
from enum import Enum
from typing import Any, Dict, List, Optional, Tuple, Union

import pytest

from typed_settings.attrs import option, secret, settings


class LeEnum(Enum):
    spam = "Le Spam"
    eggs = "Le Eggs"


@settings
class S:
    u: str = option()
    p: str = secret()


class TestAttrExtensions:
    """Tests for attrs extensions."""

    @pytest.fixture
    def inst(self):
        return S(u="spam", p="42")

    def test_secret_str(self, inst):
        assert str(inst) == "S(u='spam', p=***)"

    def test_secret_repr(self, inst):
        assert repr(inst) == "S(u='spam', p=***)"


@pytest.mark.parametrize(
    "typ, value, expected",
    [
        # Bools can be parsed from a defined set of values
        (bool, True, True),
        (bool, "True", True),
        (bool, "true", True),
        (bool, "yes", True),
        (bool, "1", True),
        (bool, 1, True),
        (bool, False, False),
        (bool, "False", False),
        (bool, "false", False),
        (bool, "no", False),
        (bool, "0", False),
        (bool, 0, False),
        # Other simple types
        (int, 23, 23),
        (int, "42", 42),
        (float, 3.14, 3.14),
        (float, ".815", 0.815),
        (str, "spam", "spam"),
        (datetime, "2020-05-04T13:37:00", datetime(2020, 5, 4, 13, 37)),
        # Enums are parsed from their "value"
        (LeEnum, "Le Eggs", LeEnum.eggs),
        # (Nested) attrs classes
        (S, {"u": "user", "p": "pwd"}, S("user", "pwd")),
        # Container types
        (List[int], [1, 2], [1, 2]),
        (List[S], [{"u": 1, "p": 2}], [S("1", "2")]),
        (Dict[str, int], {"a": 1, "b": 3.1}, {"a": 1, "b": 3}),
        (Dict[str, S], {"a": {"u": "u", "p": "p"}}, {"a": S("u", "p")}),
        (Tuple[str, ...], [1, "2", 3], ("1", "2", "3")),
        (Tuple[int, bool, str], [0, "0", 0], (0, False, "0")),
        # "Special types"
        (Any, 2, 2),
        (Any, "2", "2"),
        (Any, None, None),
        (Optional[str], 1, "1"),
        (Optional[S], None, None),
        (Optional[S], {"u": "u", "p": "p"}, S("u", "p")),
        (Optional[LeEnum], "Le Spam", LeEnum.spam),
        (Union[None, S, List[str]], None, None),
        (Union[None, S, List[str]], {"u": "u", "p": "p"}, S("u", "p")),
        (Union[None, S, List[str]], [1, 2], ["1", "2"]),
    ],
)
def test_supported_types(typ, value, expected):
    """
    All oficially supported types can be converted by attrs.

    Please create an issue if something is missing here.
    """

    @settings
    class S:
        opt: typ

    assert S(value).opt == expected



"""
Tests for `typed_settings.attrs.converters`.
"""
import sys
from datetime import datetime, timedelta, timezone

import attr
import pytest

from typed_settings.attrs.converters import (
    to_attrs,
    to_bool,
    to_dt,
    to_iterable,
    to_mapping,
    to_tuple,
    to_union,
)


class TestToAttrs:
    """Tests for `to_attrs`."""

    def test_from_data(self):
        """
        Dicts can be converted to class instances.
        """

        @attr.s
        class C:
            x = attr.ib()
            y = attr.ib()

        converter = to_attrs(C)
        assert converter({"x": 2, "y": 3}) == C(2, 3)

    def test_from_inst(self):
        """
        Existing instances remain unchanged.
        """

        @attr.s
        class C:
            x = attr.ib()
            y = attr.ib()

        inst = C(2, 3)
        converter = to_attrs(C)
        assert converter(inst) is inst

    @pytest.mark.skipif(
        sys.version_info < (3, 6),
        reason="__init_subclass__ is not yet supported",
    )
    def test_from_dict_factory(self):
        """
        Classes can specify a "from_dict" factory that will be called.
        """

        @attr.s
        class Animal:
            type = attr.ib()
            __classes__ = {}

            def __init_subclass__(cls, **kwargs):
                super().__init_subclass__(**kwargs)
                cls.__classes__[cls.__name__] = cls

            @classmethod
            def from_dict(cls, **attribs):
                cls_name = attribs["type"]
                return cls.__classes__[cls_name](**attribs)

        @attr.s(kw_only=True)
        class Cat(Animal):
            x = attr.ib()

        @attr.s(kw_only=True)
        class Dog(Animal):
            x = attr.ib()
            y = attr.ib(default=3)

        converter = to_attrs(Animal)
        assert converter({"type": "Cat", "x": 2}) == Cat(type="Cat", x=2)
        assert converter({"type": "Dog", "x": 2}) == Dog(type="Dog", x=2, y=3)

    def test_invalid_cls(self):
        """
        Raise TypeError when neither a dict nor an instance of the class is
        passed.
        """

        @attr.s
        class C:
            x = attr.ib()
            y = attr.ib()

        converter = to_attrs(C)
        with pytest.raises(TypeError):
            converter([2, 3])


class TestToDt:
    """Tests for `to_dt`."""

    def test_from_dt(self):
        """
        Existing datetimes are returned unchanged.
        """
        dt = datetime(2020, 5, 4, 13, 37)
        result = to_dt(dt)
        assert result is dt

    @pytest.mark.parametrize(
        "input, expected",
        [
            ("2020-05-04 13:37:00", datetime(2020, 5, 4, 13, 37)),
            ("2020-05-04T13:37:00", datetime(2020, 5, 4, 13, 37)),
            (
                "2020-05-04T13:37:00Z",
                datetime(2020, 5, 4, 13, 37, tzinfo=timezone.utc),
            ),
            (
                "2020-05-04T13:37:00+00:00",
                datetime(2020, 5, 4, 13, 37, tzinfo=timezone.utc),
            ),
            (
                "2020-05-04T13:37:00+02:00",
                datetime(
                    2020,
                    5,
                    4,
                    13,
                    37,
                    tzinfo=timezone(timedelta(seconds=7200)),
                ),
            ),
        ],
    )
    def test_from_str(self, input, expected):
        """
        Existing datetimes are returned unchanged.
        """
        result = to_dt(input)
        assert result == expected

    def test_invalid_input(self):
        """
        Invalid inputs raises a TypeError.
        """
        with pytest.raises(TypeError):
            to_dt(3)


class TestToBool:
    """Tests for `to_bool`."""

    @pytest.mark.parametrize(
        "val, expected",
        [
            (True, True),
            ("True", True),
            ("true", True),
            ("yes", True),
            ("1", True),
            (1, True),
            (False, False),
            ("False", False),
            ("false", False),
            ("no", False),
            ("0", False),
            (0, False),
        ],
    )
    def test_to_bool(self, val, expected):
        """
        Only a limited set of values can be converted to a bool.
        """
        assert to_bool(val) is expected

    @pytest.mark.parametrize("val", ["", [], "spam", 2, -1])
    def test_to_bool_error(self, val):
        """
        In contrast to ``bool()``, `to_bool` does no take Pythons default
        truthyness into account.

        Everything that is not in the sets above raises an error.
        """
        pytest.raises(ValueError, to_bool, val)


class TestToIterable:
    """Tests for `to_iterable`."""

    @pytest.mark.parametrize("cls", [list, set, tuple])
    def test_to_iterable(self, cls):
        """
        An iterable's data and the iterable itself can be converted to
        different types.
        """
        converter = to_iterable(cls, int)
        assert converter(["1", "2", "3"]) == cls([1, 2, 3])


class TestToTuple:
    """Tests for `to_tuple`."""

    @pytest.mark.parametrize("cls", [tuple])
    def test_to_tuple(self, cls):
        """
        Struct-like tuples can contain different data types.
        """
        converter = to_tuple(cls, [int, float, str])
        assert converter(["1", "2.2", "s"]) == cls([1, 2.2, "s"])

    @pytest.mark.parametrize("val", [["1", "2.2", "s"], ["1"]])
    def test_tuple_wrong_input_length(self, val):
        """
        Input data must have exactly as many elements as the tuple definition
        has converters.
        """
        converter = to_tuple(tuple, [int, float])
        with pytest.raises(
            TypeError,
            match="Value must have 2 items but has: {}".format(len(val)),
        ):
            converter(val)


class TestToMapping:
    """Tests for `to_mapping`."""

    @pytest.mark.parametrize("cls", [dict])
    def test_to_dict(self, cls):
        """
        Keys and values of dicts can be converted to (different) types.
        """
        converter = to_mapping(cls, int, float)
        assert converter({"1": "2", "2": "2.5"}) == cls([(1, 2.0), (2, 2.5)])


class TestToUnion:
    """Tests for `to_union`."""

    @pytest.mark.parametrize(
        "types, val, expected_type, expected_val",
        [
            ([type(None), int], None, type(None), None),
            ([type(None), int], "3", int, 3),
            ([int, float], "3", int, 3),
            ([int, float], 3.2, float, 3.2),  # Do not cast 3.2 to int!
            ([int, float], "3.2", float, 3.2),
            ([int, float, str], "3.2s", str, "3.2s"),
            ([int, float, bool, str], "3.2", str, "3.2"),
            ([int, float, bool, str], True, bool, True),
            ([int, float, bool, str], "True", str, "True"),
            ([int, float, bool, str], "", str, ""),
        ],
    )
    def test_to_union(self, types, val, expected_type, expected_val):
        """
        Union data is converted to the first matching type.  If the input data
        already has a valid type, it is returned without conversion.  For
        example, floats will not be converted to ints when the type is
        "Union[int, float]".
        """
        converter = to_union(types)
        result = converter(val)
        assert type(result) is expected_type
        assert result == expected_val



from enum import Enum
from functools import update_wrapper
from pathlib import Path
from typing import Any, Callable, Iterable, Optional, Type, Union

import attr
import click

from ._core import AUTO, T, _Auto, _load_settings
from ._dict_utils import _deep_fields, _get_path, _merge_dicts, _set_path


AnyFunc = Callable[..., Any]
Decorator = Callable[[AnyFunc], AnyFunc]


class EnumChoice(click.Choice):
    """*Click* parameter type for representing enums."""

    def __init__(self, enum_type: Type[Enum]):
        self.__enum = enum_type
        super().__init__(enum_type.__members__)

    def convert(
        self,
        value: str,
        param: Optional[click.Parameter],
        ctx: Optional[click.Context],
    ) -> Enum:
        return self.__enum[super().convert(value, param, ctx)]


def click_options(
    cls: Type[T],
    appname: str,
    config_files: Iterable[Union[str, Path]] = (),
    config_file_section: Union[_Auto, str] = AUTO,
    config_files_var: Union[None, _Auto, str] = AUTO,
    env_prefix: Union[None, _Auto, str] = AUTO,
) -> Callable[[Callable], Callable]:
    """
    Generates :mod:`click` options for a CLI which override settins loaded via
    :func:`.load_settings()`.

    A single *cls* instance is passed to the decorated function

    Example:

      .. code-block:: python

         >>> import click
         >>> import typed_settings as ts
         >>>
         >>> @ts.settings
         ... class Settings: ...
         ...
         >>> @click.command()
         ... @ts.click_options(Settings, "example")
         ... def cli(settings):
         ...     print(settings)

    See :func:`.load_settings()` for argument descriptions.
    """
    cls = attr.resolve_types(cls)
    fields = _deep_fields(cls)
    settings = _load_settings(
        fields=fields,
        appname=appname,
        config_files=config_files,
        config_file_section=config_file_section,
        config_files_var=config_files_var,
        env_prefix=env_prefix,
    )

    def pass_settings(f: AnyFunc) -> Decorator:
        """
        Creates a *cls* instances from the settings dict stored in
        :attr:`click.Context.obj` and passes it to the decorated function *f*.
        """

        def new_func(*args, **kwargs):
            ctx = click.get_current_context()
            try:
                _merge_dicts(settings, ctx.obj.get("settings"))
                ctx.obj["settings"] = cls(**settings)
            except (
                AttributeError,
                FileNotFoundError,
                TypeError,
                ValueError,
            ) as e:
                raise click.ClickException(e)
            return f(ctx.obj["settings"], *args, **kwargs)

        return update_wrapper(new_func, f)

    def wrap(f):
        for path, field, _cls in reversed(fields):
            try:
                default = _get_path(settings, path)
            except KeyError:
                default = field.default
            option = _mk_option(click.option, path, field, default)
            f = option(f)
        f = pass_settings(f)
        return f

    return wrap


def pass_settings(f: AnyFunc) -> AnyFunc:
    """
    Marks a callback as wanting to receive the innermost settings instance as
    first argument.
    """

    def new_func(*args, **kwargs):
        ctx = click.get_current_context()
        node = ctx
        settings = None
        while node is not None:
            if isinstance(node.obj, dict) and "settings" in node.obj:
                settings = node.obj["settings"]
                break
            node = node.parent
        return ctx.invoke(f, settings, *args, **kwargs)

    return update_wrapper(new_func, f)


def _mk_option(
    option: Callable[..., Decorator],
    path: str,
    field: attr.Attribute,
    default: Any,
) -> Decorator:
    """Recursively creates click options and returns them as a list."""

    def cb(ctx, _param, value):
        if ctx.obj is None:
            ctx.obj = {}
        settings = ctx.obj.setdefault("settings", {})
        _set_path(settings, path, value)
        return value

    kwargs = {}
    if default is not attr.NOTHING:
        kwargs["default"] = default

    opt_name = path.replace(".", "-").replace("_", "-")
    param_decl = f"--{opt_name}"
    option_type = field.type
    if field.type is bool:
        param_decl = f"{param_decl}/--no-{opt_name}"
    if field.type and issubclass(field.type, Enum):
        option_type = EnumChoice(field.type)  # type: ignore
        if "default" in kwargs:
            kwargs["default"] = kwargs["default"].name

    return option(
        param_decl,
        type=option_type,
        show_default=True,
        callback=cb,
        expose_value=False,
        is_eager=True,
        **kwargs,
    )



"""
Typed settings
"""
from typing import Any, List

from ._core import load_settings, update_settings
from .attrs import option, secret, settings


__all__ = [
    "click_options",
    "load_settings",
    "option",
    "pass_settings",
    "secret",
    "settings",
    "update_settings",
]


def __getattr__(name: str) -> Any:
    if name == "click_options":
        from ._click import click_options

        return click_options
    if name == "pass_settings":
        from ._click import pass_settings

        return pass_settings

    raise AttributeError(f"module '{__name__}' has no attribute '{name}'")


def __dir__() -> List[str]:
    return __all__



"""
Additional attrs validators
"""
import operator

from attr import attrib, attrs


__all__ = [
    "lt",
    "le",
    "ge",
    "gt",
    "maxlen",
]


@attrs(repr=False, frozen=True, slots=True)
class _NumberValidator:
    bound = attrib()
    compare_op = attrib()
    compare_func = attrib()

    def __call__(self, inst, attr, value):
        """
        We use a callable class to be able to change the ``__repr__``.
        """
        if not self.compare_func(value, self.bound):
            raise ValueError(
                "'{name}' must be {op} {bound}: {value}".format(
                    name=attr.name,
                    op=self.compare_op,
                    bound=self.bound,
                    value=value,
                )
            )

    def __repr__(self):
        return "<Validator for x {op} {bound}>".format(
            op=self.compare_op, bound=self.bound
        )


def lt(val):
    """
    A validator that raises `ValueError` if the initializer is called
    with a number larger or equal to *val*.

    :param int val: Exclusive upper bound for values

    .. versionadded:: 20.3.0
    """
    return _NumberValidator(val, "<", operator.lt)


def le(val):
    """
    A validator that raises `ValueError` if the initializer is called
    with a number greater than *val*.

    :param int val: Inclusive upper bound for values

    .. versionadded:: 20.3.0
    """
    return _NumberValidator(val, "<=", operator.le)


def ge(val):
    """
    A validator that raises `ValueError` if the initializer is called
    with a number smaller than *val*.

    :param int val: Inclusive lower bound for values

    .. versionadded:: 20.3.0
    """
    return _NumberValidator(val, ">=", operator.ge)


def gt(val):
    """
    A validator that raises `ValueError` if the initializer is called
    with a number smaller or equal to *val*.

    :param int val: Exclusive lower bound for values

    .. versionadded:: 20.3.0
    """
    return _NumberValidator(val, ">", operator.gt)


@attrs(repr=False, frozen=True, slots=True)
class _MaxLengthValidator:
    max_length = attrib()

    def __call__(self, inst, attr, value):
        """
        We use a callable class to be able to change the ``__repr__``.
        """
        if len(value) > self.max_length:
            raise ValueError(
                "Length of '{name}' must be <= {max}: {len}".format(
                    name=attr.name, max=self.max_length, len=len(value)
                )
            )

    def __repr__(self):
        return "<maxlen validator for {max}>".format(max=self.max_length)


def maxlen(length):
    """
    A validator that raises `ValueError` if the initializer is called
    with a string or iterable that is longer than *length*.

    :param int length: Maximum length of the string or iterable

    .. versionadded:: 20.3.0
    """
    return _MaxLengthValidator(length)



from enum import Enum
from pathlib import Path

import click
import click.testing
import pytest

from typed_settings import (
    click_options,
    option,
    pass_settings,
    secret,
    settings,
)


class LeEnum(Enum):
    spam = "le spam"
    eggs = "Le eggs"


@settings
class Nested:
    a: str = "nested"
    b: int = option(default=0, converter=int)  # type: ignore


@settings(kw_only=True)
class Settings:
    a: str
    b: str = secret()
    c: str = option(default="spam")
    d: int = 0
    e: float = 0
    f: bool = False
    g: bool = True
    h: LeEnum = option(  # type: ignore
        default=LeEnum.spam,
        converter=lambda v: LeEnum(v) if isinstance(v, str) else v,  # type: ignore  # noqa
    )
    p: Path = Path("/")
    n: Nested = option(  # type: ignore
        default=Nested(),  # type: ignore
        converter=lambda d: Nested(**d) if isinstance(d, dict) else d,  # type: ignore  # noqa
    )
    long_name: str = option(default="val")  # type: ignore


@pytest.fixture
def cli(tmp_path):
    """
    Creates a click command for ``Settings`` and returns a functions that
    invokes a click test runner with the passed arguments.

    The result object will habe a ``settings`` attribute that holds the
    generated ``Settings`` instance for verification.

    """

    class Runner(click.testing.CliRunner):
        settings: object

        def invoke(self, *args, **kwargs):
            result = super().invoke(*args, **kwargs)
            try:
                result.settings = self.settings
            except AttributeError:
                result.settings = None
            return result

    runner = Runner()

    @click.group(invoke_without_command=True)
    @click_options(Settings, "test", [tmp_path.joinpath("settings.toml")])
    def cli(settings):
        runner.settings = settings

    def run(*args, **kwargs):
        return runner.invoke(cli, args, **kwargs)

    return run


class TestClickOptions:
    """Tests for click_options()."""

    def test_click_options(self, cli):
        """Make sure click options of various types work."""

        result = cli(
            "--a=eggs1",
            "--b=eggs2",
            "--c=eggs3",
            "--d=3",
            "--e=3.1",
            "--f",
            "--no-g",
            "--h=eggs",
            "--p=/spam",
            "--n-a=eggs4",
            "--n-b=3",
        )
        assert result.output == ""
        assert result.exit_code == 0
        assert result.settings == Settings(
            a="eggs1",
            b="eggs2",
            c="eggs3",
            d=3,
            e=3.1,
            f=True,
            g=False,
            h=LeEnum.eggs,
            p=Path("/spam"),
            n=Nested("eggs4", 3),
        )

    def test_click_defaults(self, cli):
        """
        Most types do not accept "None" as value, so most options need to have
        a default value.
        """
        result = cli()
        assert result.output == ""
        assert result.exit_code == 0
        assert result.settings == Settings(
            a=None,
            b=None,
        )

    def test_help(self, cli):
        """All options get a proper help string."""
        result = cli("--help")
        assert result.output == (
            "Usage: cli [OPTIONS] COMMAND [ARGS]...\n"
            "\n"
            "Options:\n"
            "  --a TEXT\n"
            "  --b TEXT\n"
            "  --c TEXT          [default: spam]\n"
            "  --d INTEGER       [default: 0]\n"
            "  --e FLOAT         [default: 0]\n"
            "  --f / --no-f      [default: False]\n"
            "  --g / --no-g      [default: True]\n"
            "  --h [spam|eggs]   [default: spam]\n"
            "  --p PATH          [default: /]\n"
            "  --n-a TEXT        [default: nested]\n"
            "  --n-b INTEGER     [default: 0]\n"
            "  --long-name TEXT  [default: val]\n"
            "  --help            Show this message and exit.\n"
        )
        assert result.exit_code == 0

    def test_click_default_from_settings(self, monkeypatch, tmp_path):
        """
        If a setting is set in a config file, that value is being used as
        default for click options - *not* the default defined in the Settings
        class.
        """

        tmp_path.joinpath("settings.toml").write_text('[test]\na = "x"\n')
        spath = tmp_path.joinpath("settings2.toml")
        print(spath)
        spath.write_text('[test]\nb = "y"\n')
        monkeypatch.setenv("TEST_SETTINGS", str(spath))
        monkeypatch.setenv("TEST_C", "z")

        @settings
        class Settings:
            a: str
            b: str
            c: str
            d: str

        @click.command()
        @click_options(Settings, "test", [tmp_path.joinpath("settings.toml")])
        def cli(settings):
            print(settings)

        runner = click.testing.CliRunner()
        result = runner.invoke(cli, ["--help"])
        assert result.output == (
            "Usage: cli [OPTIONS]\n"
            "\n"
            "Options:\n"
            "  --a TEXT  [default: x]\n"
            "  --b TEXT  [default: y]\n"
            "  --c TEXT  [default: z]\n"
            "  --d TEXT\n"
            "  --help    Show this message and exit.\n"
        )
        assert result.exit_code == 0


class TestPassSettings:
    """Tests for pass_settings()."""

    @settings
    class Settings:
        opt: str = ""

    def test_pass_settings(self):
        """
        A subcommand can receive the settings via the `pass_settings`
        decorator.
        """

        @click.group()
        @click_options(self.Settings, "test")
        def cli(settings):
            pass

        @cli.command()
        @pass_settings
        def cmd(settings):
            print(settings)
            assert settings == self.Settings(opt="spam")

        runner = click.testing.CliRunner()
        result = runner.invoke(cli, ["--opt=spam", "cmd"])
        assert result.output == "TestPassSettings.Settings(opt='spam')\n"
        assert result.exit_code == 0

    def test_pass_settings_no_settings(self):
        """
        Pass ``None`` if no settings are defined.
        """

        @click.group()
        def cli():
            pass

        @cli.command()
        @pass_settings
        def cmd(settings):
            print(settings)
            assert settings is None

        runner = click.testing.CliRunner()
        result = runner.invoke(cli, ["cmd"])
        assert result.output == "None\n"
        assert result.exit_code == 0



"""
Helpers for and additions to attrs.
"""
from datetime import datetime
from functools import partial

import attr

from .converters import to_bool, to_dt
from .hooks import make_auto_converter


class _SecretRepr:
    def __call__(self, _v):
        return "***"

    def __repr__(self):
        return "***"


auto_convert = make_auto_converter({bool: to_bool, datetime: to_dt})
"""Auto-convert supported types."""

settings = partial(attr.frozen, field_transformer=auto_convert)
"""An alias to :func:`attr.frozen()`"""

# settings = partial(attr.frozen, field_transformer=attr.auto_convert)
option = partial(attr.field)
"""An alias to :func:`attr.field()`"""

secret = partial(attr.field, repr=_SecretRepr())
"""
An alias to :func:`attr.field()`.

When printing a settings instances, secret settings will represented with `***`
istead of their actual value.

Example:

  .. code-block:: python

     >>> from typed_settings import settings, secret
     >>>
     >>> @settings
     ... class Settings:
     ...     password: str = secret()
     ...
     >>> Settings(password="1234")
     Settings(password=***)
"""



"""
Tests for `typed_settings.attrs.validators`.
"""
import attr
import pytest
from attr import fields

from typed_settings.attrs import validators as validator_module
from typed_settings.attrs.validators import ge, gt, le, lt, maxlen


class TestLtLeGeGt:
    """
    Tests for `maxlen`.
    """

    BOUND = 4

    def test_in_all(self):
        """
        validator is in ``__all__``.
        """
        assert all(
            f.__name__ in validator_module.__all__ for f in [lt, le, ge, gt]
        )

    @pytest.mark.parametrize("v", [lt, le, ge, gt])
    def test_retrieve_bound(self, v):
        """
        The configured bound for the comparison can be extracted from the
        Attribute.
        """

        @attr.s
        class Tester:
            value = attr.ib(validator=v(self.BOUND))

        assert fields(Tester).value.validator.bound == self.BOUND

    @pytest.mark.parametrize(
        "v, value",
        [
            (lt, 3),
            (le, 3),
            (le, 4),
            (ge, 4),
            (ge, 5),
            (gt, 5),
        ],
    )
    def test_check_valid(self, v, value):
        """Silent if value {op} bound."""

        @attr.s
        class Tester:
            value = attr.ib(validator=v(self.BOUND))

        Tester(value)  # shouldn't raise exceptions

    @pytest.mark.parametrize(
        "v, value",
        [
            (lt, 4),
            (le, 5),
            (ge, 3),
            (gt, 4),
        ],
    )
    def test_check_invalid(self, v, value):
        """Raise ValueError if value {op} bound."""

        @attr.s
        class Tester:
            value = attr.ib(validator=v(self.BOUND))

        with pytest.raises(ValueError):
            Tester(value)

    @pytest.mark.parametrize("v", [lt, le, ge, gt])
    def test_repr(self, v):
        """
        __repr__ is meaningful.
        """
        nv = v(23)
        assert repr(nv) == "<Validator for x {op} {bound}>".format(
            op=nv.compare_op, bound=23
        )


class TestMaxlen:
    """
    Tests for `maxlen`.
    """

    MAX_LENGTH = 4

    def test_in_all(self):
        """
        validator is in ``__all__``.
        """
        assert maxlen.__name__ in validator_module.__all__

    def test_retrieve_maxlen(self):
        """
        The configured max. length can be extracted from the Attribute
        """

        @attr.s
        class Tester:
            value = attr.ib(validator=maxlen(self.MAX_LENGTH))

        assert fields(Tester).value.validator.max_length == self.MAX_LENGTH

    @pytest.mark.parametrize(
        "value",
        [
            "",
            "foo",
            "spam",
            [],
            list(range(MAX_LENGTH)),
            {"spam": 3, "eggs": 4},
        ],
    )
    def test_check_valid(self, value):
        """
        Silent if len(value) <= maxlen.
        Values can be strings and other iterables.
        """

        @attr.s
        class Tester:
            value = attr.ib(validator=maxlen(self.MAX_LENGTH))

        Tester(value)  # shouldn't raise exceptions

    @pytest.mark.parametrize(
        "value",
        [
            "bacon",
            list(range(6)),
        ],
    )
    def test_check_invalid(self, value):
        """
        Raise ValueError if len(value) > maxlen.
        """

        @attr.s
        class Tester:
            value = attr.ib(validator=maxlen(self.MAX_LENGTH))

        with pytest.raises(ValueError):
            Tester(value)

    def test_repr(self):
        """
        __repr__ is meaningful.
        """
        assert repr(maxlen(23)) == "<maxlen validator for 23>"



"""
Addtional attrs hooks
"""
from datetime import datetime
from enum import Enum
from typing import Any, TypeVar, Union, get_type_hints

from ._compat import get_args, get_origin
from .converters import (
    to_attrs,
    to_dt,
    to_iterable,
    to_mapping,
    to_tuple,
    to_union,
)


__all__ = [
    "auto_convert",
    "auto_serialize",
    "make_auto_converter",
]


def make_auto_converter(converters):
    """
    Return a

    """

    def auto_convert(cls, attribs):
        """
        A field transformer that tries to convert all attribs of a class to
        their annotated type.
        """
        # We cannot use attrs.resolve_types() here,
        # because "cls" is not yet a finished attrs class:
        type_hints = get_type_hints(cls)
        results = []
        for attrib in attribs:
            # Do not override explicitly defined converters!
            if attrib.converter is None:
                converter = _get_converter(type_hints[attrib.name], converters)
                attrib = attrib.evolve(converter=converter)
            results.append(attrib)

        return results

    return auto_convert


def _get_converter(
    type_,
    converters,
    iterable_types=frozenset({list, set, frozenset}),
    tuple_types=frozenset({tuple}),
    mapping_types=frozenset({dict}),
):
    """
    Recursively resolves concrete and generic types and return a proper
    converter.
    """
    # Detect whether "type_" is a container type.  Currently we need
    # to check, e.g., for "typing.List".  From python 3.9, we also
    # need to check for "list" directly.
    origin = get_origin(type_)
    if origin is None:
        converter = _handle_concrete(type_, converters)
    else:
        converter = _handle_generic(
            type_, converters, iterable_types, tuple_types, mapping_types
        )

    return converter


def _handle_concrete(type_, converters):
    """
    Returns a converter for concrete types.

    These include attrs classes, :code:`Any` andall types in *converters*.
    """
    # Get converter for concrete type
    if type_ is Any:
        converter = _to_any
    elif getattr(type_, "__attrs_attrs__", None) is not None:
        # Attrs classes
        converter = to_attrs(type_)
    else:
        # Check if type is in converters dict
        for convert_type, convert_func in converters.items():
            if issubclass(type_, convert_type):
                converter = convert_func
                break
        else:
            # Fall back to simple types like bool, int, str, Enum, ...
            converter = type_
    return converter


def _handle_generic(
    type_, converters, iterable_types, tuple_types, mapping_types
):
    """
    Returns a converter for generic types like lists, tuples, dicts or Union.
    """
    origin = get_origin(type_)
    args = get_args(type_)
    if origin in iterable_types:
        item_converter = _get_converter(args[0], converters)
        converter = to_iterable(origin, item_converter)
    elif origin in tuple_types:
        if len(args) == 2 and args[1] == ...:
            # "frozen list" variant of tuple
            item_converter = _get_converter(args[0], converters)
            converter = to_iterable(tuple, item_converter)
        else:
            # "struct" variant of tuple
            item_converters = [_get_converter(t, converters) for t in args]
            converter = to_tuple(tuple, item_converters)
    elif origin in mapping_types:
        key_converter = _get_converter(args[0], converters)
        val_converter = _get_converter(args[1], converters)
        converter = to_mapping(dict, key_converter, val_converter)
    elif origin is Union:
        item_converters = [_get_converter(t, converters) for t in args]
        converter = to_union(item_converters)
    else:
        raise TypeError(f"Cannot create converter for generic type: {type_}")
    return converter


A = TypeVar("A")


def _to_any(val: A) -> A:
    return val


# TODO: Also add "to_bool()"?
auto_convert = make_auto_converter({datetime: to_dt})
"""Auto-convert :class:`datetime.datetime` as well as other stuff."""


def auto_serialize(_inst, _attrib, value):
    """Inverse hook to :func:`auto_convert` for use with
    :func:`attrs.asdict()`.
    """
    if isinstance(value, datetime):
        return datetime.isoformat(value)
    if isinstance(value, Enum):
        return value.value
    return value
